/**
    This is "qwz" (pronounced: "cues" or "quiz"), 
    the "Query Wizard for the MySQL command line"
    Copyright (C) 2008 Roland Bouman http://rpbouman.blogspot.com/

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
set @old_sql_mode := @@sql_mode;

set @@sql_mode := 'PIPES_AS_CONCAT,ANSI_QUOTES,IGNORE_SPACE,NO_AUTO_VALUE_ON_ZERO,STRICT_TRANS_TABLES,STRICT_ALL_TABLES,NO_ZERO_IN_DATE,NO_ZERO_DATE,ERROR_FOR_DIVISION_BY_ZERO,TRADITIONAL,NO_AUTO_CREATE_USER';

create schema if not exists qwz;

use qwz;

set @qwz_state := null;

drop temporary table if exists qwz_menu;
drop temporary table if exists qwz_join_condition;
drop temporary table if exists qwz_table;
drop temporary table if exists qwz_table_ancestor;
drop temporary table if exists qwz_column;
drop temporary table if exists qwz_join_column;
drop temporary table if exists qwz_menu_column1;
drop temporary table if exists qwz_menu_column2;
drop temporary table if exists qwz_sql;

delimiter //

drop procedure if exists qwz
//

create procedure qwz(
    p_choice varchar(64)
)
sql security invoker
comment 'Interactive Query Wizard for the MySQL command line'
_main: begin
    declare v_menu_groups text;
    declare v_menu_group enum(
        'choose columns'
    ,   'choose custom join'
    ,   'choose execute sql'
    ,   'choose explain sql'
    ,   'choose group by columns'
    ,   'choose help'
    ,   'choose show sql'
    ,   'choose schema'
    ,   'choose related tables'
    ,   'choose tables'
    ,   'columns'
    ,   'edit join condition'
    ,   'execute sql'
    ,   'explain sql'
    ,   'help'
    ,   'new query'
    ,   'related tables'
    ,   'schemata'
    ,   'set schema'
    ,   'show sql'
    ,   'start'
    ,   'tables'
    );
    declare v_states set(
        'columns'
    ,   'custom join'
    ,   'group by columns'
    ,   'help'
    ,   'execute sql'
    ,   'explain sql'
    ,   'order by columns'
    ,   'related tables'
    ,   'schemata'
    ,   'set column'
    ,   'set group by column'
    ,   'set join condition'
    ,   'set order by column'
    ,   'set related table'
    ,   'set schema'
    ,   'set table'
    ,   'show sql'
    ,   'start'
    ,   'tables'
    ,   'unset related table'
    );

    declare v_bell                 char(1) 
        default char(7);
    declare v_label_alias          varchar(255) 
        default 'The command line Query Wizard for MySQL';

    declare v_id                   int unsigned;
    declare v_left_id              int unsigned;
    declare v_right_id             int unsigned;

    declare v_schema_name          varchar(64);
    declare v_table_name           varchar(64);
    declare v_included             enum('YES','NO');

    declare v_join_type            enum('CROSS', 'INNER','LEFT', 'RIGHT');
    declare v_choice               varchar(64);
    declare v_value                varchar(64);
    declare v_label                text;
    declare v_count_states         int unsigned;
    declare v_count_choices        int unsigned;
    declare v_count_related_tables int unsigned;

    declare v_done                 bool default FALSE;
    declare csr_tables cursor for
        select   id
        ,        schema_name, table_name
        ,        join_type, included
        from     qwz_table
        where    find_in_set(id,v_value)
        order by if(included='YES',1,2)
        ,        if(included='YES',right_id,left_id);

    _create_tables: begin
    /**
    *   This block creates all temporary tables
    *   It would be possible to create all these 
    *   as persistent tables outside this proc, 
    *   but currently we favour a stand alone,
    *   and self-contained stored proc.
    */
        -- warning thrown when a table does not exist
        declare cond_table_exists condition for 1050;
        -- ignore warnings about non-existent tables
        declare continue handler for cond_table_exists begin end;

        -- this table is what we use to present a menu to the user
        -- for the user, the position is the number to be entered 
        -- to pick an option, and the label is the accompanying text.
        -- The state is the state to plunge into 
        -- when the option is chosen.
        -- The value is - if appplicable - the variable chosen value.
        -- The value is usually an identifier - table name etc.
        create temporary table if not exists qwz_menu(
            id       int unsigned     not null auto_increment
        ,   position varchar(10)      not null
        ,   label    varchar(255)     not null
        ,   state    varchar(64)      not null
        ,   value    varchar(64)      not null default ''
        ,   visible  enum('YES','NO') not null default 'YES'
        ,   primary key(id)
        ) engine = memory;

        -- Nested set of tables reprents the join path
        -- left_id, right_id: maintain a nested set
        -- schema_name, table_name: identifies the table
        -- constraint_schema, constraint_name: 
        --   identifies the foreign key that joins this table 
        --   the joined table is the child table in db terms,
        --   but the parent table in the nested set.
        --   i.o.w, the nested set lists 'lookup' tables
        --   as children.
        -- included defines whether the table is part of the query
        -- join_type specifies how this table is joined
        create temporary table if not exists qwz_table( 
            id                int unsigned          not null auto_increment
        ,   left_id           int unsigned          not null
        ,   right_id          int unsigned          not null
        ,   schema_name       varchar(64)           not null
        ,   table_name        varchar(64)           not null
        ,   constraint_schema varchar(64)
        ,   constraint_name   varchar(64)
        ,   included          enum('YES','NO')      not null default 'NO'
        ,   join_type         enum('CROSS'
                              ,    'INNER'
                              ,    'LEFT'
                              ,    'RIGHT')         default null
        ,   primary key(id)
        ,   unique(left_id)
        ) engine = memory;

        -- require an extract of the table nested set 
        -- in order to join them
        create temporary table if not exists qwz_table_ancestor(
            left_id         int unsigned          not null
        ,   right_id        int unsigned          not null
        ,   primary key(left_id)
        ) engine = memory;

        -- SELECT -list columns
        create temporary table if not exists qwz_column(
            id                   int unsigned auto_increment not null
        ,   table_id             int unsigned                not null
        ,   column_name          varchar(64)                 not null
        ,   ordinal_position     int unsigned                not null
        ,   position_in_select   int unsigned                         default null
        ,   position_in_order_by int unsigned                         default null
        ,   order_by_direction   enum('ASC','DESC')                   default null
        ,   position_in_group_by int unsigned                         default null
        ,   included             enum('YES','NO')            not null default 'NO'
        ,   primary key(id)
        ) engine = memory;

        create temporary table if not exists qwz_join_column(
            id                 int unsigned
        ,   table_left_id      int unsigned                not null
        ,   column_name        varchar(64)                 not null
        ,   primary key(id)
        ) engine = memory;

        create temporary table if not exists qwz_menu_column1(
            position varchar(10)      not null
        ,   value    varchar(64)      not null default ''
        ) engine = memory;

        create temporary table if not exists qwz_menu_column2
        like qwz_menu_column1;

        create temporary table if not exists qwz_join_condition(
            id                  int unsigned auto_increment not null
        ,   column_id1          int unsigned                not null
        ,   column_id2          int unsigned                not null
        ,   comparison_operator enum(
                                    '='
                                ,   '>'
                                ,   '<'
                                ,   '>='
                                ,   '<='
                                ,   '!='
                                ,   '<=>'
                                )                           not null
        ,   primary key(id)
        ,   unique(column_id1,column_id2)
        ) engine = memory;

        -- Generated SQL
        create temporary table if not exists qwz_sql(
            id               int unsigned auto_increment not null
        ,   sql_text         varchar(255)                not null
        ,   position_in_select int unsigned                       default null
        ,   left_id            int unsigned                       default null
        ,   primary key(id)
        ) engine = memory;
    end;

    _real: begin
        _init: begin
            declare cond_no_row condition for 1329; 
            declare cond_multiple_states condition for 1172;
            declare cond_unknown_state condition for 1265;

            declare continue handler for cond_no_row begin end;
            declare exit handler for cond_unknown_state 
                begin
                    select concat(group_concat(state),v_bell)  'some unknown states'
                    from   qwz_menu
                    where  find_in_set(position, p_choice);
                end;

            set @@group_concat_max_len := @@max_allowed_packet;
            set @qwz_pos := 0;
            set @qwz_id  := 0;
            if @qwz_schema is null then 
                set @qwz_schema := schema();
            end if;
            if p_choice = 'help' then
                set @qwz_state := 'help';
                leave _init;
            elseif @qwz_state is null 
            or     p_choice is null 
            or     p_choice in ('start','reset','new') then
                set @qwz_state := 'start';
                leave _init;
            end if;

            select   count(distinct state)
            ,        count(distinct position)
            ,        group_concat(distinct position order by find_in_set(position,p_choice))
            ,        group_concat(distinct value order by find_in_set(position,p_choice))
            ,        group_concat(distinct state order by find_in_set(position,p_choice))
            ,        group_concat(distinct label order by find_in_set(position,p_choice))
            into     v_count_states
            ,        v_count_choices
            ,        v_choice
            ,        v_value 
            ,        v_states
            ,        v_label
            from     qwz_menu
            where    find_in_set(position, p_choice);

            if v_count_states = 1 then
                set @qwz_state := v_states;
                truncate table qwz_menu;
            else
                if v_count_states = 0 then
                    select   concat(
                                 'Your choice (', coalesce(p_choice,'NULL'), ') is not applicable'
                             ,   ' in the current context.'
                             ,   v_bell
                             )   "Invalid selection"
                    union all 
                    select   'Valid choices are limited to the numbers that appear left in the menu.';
                elseif v_count_states > 1 then
                -- If we arrive here, the user made a selection out of more than
                -- one menu group, which we currently can't handle. 
                -- In this case, we notify the user and offer the same menu again
                -- We would like to use a HANDLER for cond_multiple_states (1172)
                -- but can't because we can't LEAVE the _real block from there
                    select   concat(
                                 'Multiple choices (', coalesce(p_choice,'NULL') ,')'
                             ,   ' from multiple menu groups (', group_concat(state),').'
                             ,   v_bell
                             )   "Invalid selection"
                    from     qwz_menu
                    where    find_in_set(position, p_choice)
                    union all
                    select 'Multiple choices must be from the same menu group';
                end if;
                leave _real;
            end if;

            if v_count_choices > 1 and @qwz_state in (
                'set schema','set table','custom join'
            ) then
                set v_value := substring_index(v_value,',',1);
                set v_label := substring_index(v_label,',',1);
                select concat(
                           'Multiple choices (', coalesce(p_choice,'NULL'), ')'
                       ,   ' not applicable for the current state (', coalesce(@qwz_state,'NULL'),').'
                       ,   v_bell
                       ) "Invalid selection"
                union all
                select concat('Ignoring all but the first choice (', v_value,') ...');
            elseif v_choice != p_choice then
                select concat(
                           'Some unrecognized choices (', coalesce(p_choice,'NULL'), ')'
                       ,   v_bell
                       ) "Invalid selection"
                union all
                select concat('Using only recognized choices (', v_choice,') ...');
            end if;

        end _init;

        /**
        *   In this block we generate SQL
        */
        if @qwz_state in ('show sql','execute sql','explain sql') then
        -- if we arrive here, we will need to generated sql somewhere later on
            if coalesce(@qwz_dirty1,true) then
                -- if we arrive here, we need to re-generate the sql because the
                -- statement changed since the last time we generated sql
                truncate table qwz_sql;

                -- generate SQL for the SELECT list, store it in the table
                insert into qwz_sql (sql_text, position_in_select)
                select     concat(
                               rpad(
                                   if((@qwz_pos:=@qwz_pos+1) = 1, 'SELECT', ',')
                               ,   8, ' '
                               )
                           ,   't', t.left_id, '.'
                           ,   c.column_name
                           )
                ,          position_in_select
                from       qwz_column  c
                inner join qwz_table   t
                on         c.table_id = t.id
                where      c.included = 'YES'
                order by   position_in_select;

                -- generate SQL for the FROM list, store it in the table
                insert 
                into       qwz_sql (sql_text, left_id)
                select     concat(
                               if( t.left_id=1
                               ,   'FROM '
                               ,   concat(
                                       t.join_type
                                   ,   ' JOIN '
                                   )
                               )
                           ,   t.schema_name, '.',  t.table_name
                           ,   ' AS t', left_id
                           ,   if(t.left_id=1,''
                               ,   concat(
                                       ' ON ('
                                   ,   group_concat(
                                            't', t.joined_table_left_id, '.'
                                       ,    kcu.column_name 
                                       order by ordinal_position
                                       )
                                   ,   ') = ('
                                   ,   group_concat(
                                           't', t.left_id, '.'
                                       ,   kcu.referenced_column_name 
                                       order by ordinal_position
                                       )
                                   ,   ')'
                                   )
                               )
                           )
                ,          t.left_id
                from       (
                           select     t.left_id
                           ,          t.schema_name
                           ,          t.table_name
                           ,          t.constraint_schema
                           ,          t.constraint_name
                           ,          t.join_type
                           ,          t.included
                           ,          max(a.left_id)       joined_table_left_id
                           from       qwz_table t
                           left join  qwz_table_ancestor a
                           on         t.left_id  > a.left_id
                           and        t.right_id < a.right_id
                           where      t.included = 'YES'
                           and        (t.left_id = 1 or t.constraint_name is not null)
                           group by   t.left_id
                           ) as t
                left join  information_schema.key_column_usage kcu
                on         t.constraint_schema = kcu.constraint_schema
                and        t.constraint_name   = kcu.constraint_name
                and        kcu.referenced_table_schema is not null
                group by   t.left_id;

                -- generate the custom joined tables 
                truncate table qwz_join_column;
                insert into qwz_join_column(id,table_left_id,column_name)
                select     c.id,t.left_id,c.column_name
                from       qwz_join_condition jc
                inner join qwz_column         c
                on         jc.column_id1 = c.id
                inner join qwz_table          t
                on         c.table_id = t.id
                ;

                insert into qwz_sql (sql_text, left_id)
                select     concat(
                               t.join_type, ' JOIN '
                           ,   t.schema_name, '.', t.table_name
                           ,   ' AS t', t.left_id
                           ,   if(jcm.table_left_id is null, '', '\nON ') 
                           ,   coalesce(group_concat(
                                   't', jcm.table_left_id, '.'
                               ,   jcm.column_name, ' '
                               ,   jc.comparison_operator
                               , ' t',t.left_id,'.',c.column_name
                               separator '\nAND '
                               ),'')
                           )
                ,          t.left_id
                from       qwz_table t
                inner join qwz_column c
                on         t.id = c.table_id
                left join  qwz_join_condition jc
                on         c.id = jc.column_id2
                left join  qwz_join_column jcm
                on         jc.column_id1 = jcm.id
                where      t.constraint_name is null
                and        t.left_id > 1
                group by   t.left_id;

                -- generate the ORDER BY clause
                set @qwz_pos := 0;
                insert into qwz_sql (sql_text)
                select     concat(
                               rpad(if((@qwz_pos:=@qwz_pos+1)=1,'ORDER BY', ','), 9,' ')
                           ,   't', t.left_id, '.', c.column_name
                           ,   ' ', c.order_by_direction)
                from       qwz_column c
                inner join qwz_table t
                on         c.table_id = t.id
                where      not position_in_order_by is null
                order by   position_in_order_by;

                set @qwz_pos:=0;
                -- mark the generated statement cache invalid
                set @qwz_dirty2 := true;
                -- mark the generated statement pieces as valid
                set @qwz_dirty1 := false; 
            end if;

            if  @qwz_state in ('execute sql','explain sql','show sql') 
            and coalesce(@qwz_dirty2,true) then

                -- if we arrive here, 
                -- we need to have the statement as a string somewhere
                -- and we determined we can't use a cached statement
                select  group_concat(
                            sql_text 
                        order by case 
                                     when position_in_select is not null then 1
                                     when left_id is not null then 2
                                     else 1000
                                 end
                        ,        position_in_select
                        ,        left_id
                        ,        id
                        separator '\n'
                        )
                into    @qwz_sql
                from    qwz_sql;

                -- mark the generated statement cache valid
                set @qwz_dirty2 := false;
            end if;
        end if;

        -- this is where we digest the current state, process the choice made by the user
        -- we also set the continuation menus, and any contiunuation state.
        case @qwz_state
            when 'columns' then
                set v_label_alias := 'Build SELECT list: + columns are added / - columns are removed';
                set v_menu_groups := 'columns,choose related tables,choose custom join,new query';
            when 'custom join' then
                set v_label_alias := concat('Select a table to ',v_value);
                set v_menu_groups := 'tables,choose schema,new query';
            when 'group by columns' then
                set v_label_alias := 'Select GROUP BY Columns';
                set v_menu_groups := 'columns,choose related tables,choose custom join,choose order by,new query';
            when 'help' then
                set v_label_alias := 'Command line Query Wizard from MySQL - Help';
                set v_menu_groups := 'help,new query';
            when 'execute sql' then
                set v_menu_groups := 'choose related tables,choose custom join,choose columns,choose explain sql,choose show sql,new query';
                prepare qwz_sql from @qwz_sql;
                execute qwz_sql;
                deallocate prepare qwz_sql;
            when 'explain sql' then
                set v_menu_groups := 'choose related tables,choose custom join,choose columns,choose execute sql,choose show sql,new query';
                set @qwz_explain_sql:= concat('EXPLAIN ',@qwz_sql);
                prepare qwz_explain_sql from @qwz_explain_sql;
                execute qwz_explain_sql;
                deallocate prepare qwz_explain_sql;
            when 'order by columns' then
                set v_label_alias := 'Select ORDER BY Columns';
                set v_menu_groups := 'columns,choose related tables,choose custom join,new query';
            when 'related tables' then
                set v_label_alias := 'Add a joined table';
                set v_menu_groups := 'related tables,choose custom join,choose columns,new query';
            when 'start' then
                set v_label_alias := 'QWZ - a command line Query Wizard for MySQL!!! Type: CALL qwz(''help''); to get help.';
                set v_menu_groups := 'choose schema,choose tables,choose help';
                truncate table qwz_column;
                truncate table qwz_join_column;
                truncate table qwz_table;
                truncate table qwz_table_ancestor;
                truncate table qwz_join_condition;
                truncate table qwz_menu_column1;
                truncate table qwz_menu_column2;
                truncate table qwz_sql;
            when 'schemata' then
                set v_label_alias := 'Set the default schema';
                set v_menu_groups := 'schemata';
            when 'set column' then
                set v_label_alias := 'Build SELECT list: + columns are added / - columns are removed';
                set v_menu_groups := 'columns,choose related tables,choose custom join,choose show sql,choose explain sql,choose execute sql,new query';

                update qwz_column
                set    position_in_select = if(
                           included='YES',null
                       ,   find_in_set(id,v_choice)
                       )
                ,      included = if(included='YES','NO','YES')
                where  find_in_set(id,v_choice);

                set @qwz_dirty1 := true;
            when 'set order by column' then
                set v_menu_groups := 'columns,choose related tables,choose custom join,choose show sql,choose explain sql,choose execute sql,new query';
                update     qwz_column
                set        position_in_order_by = find_in_set(id,v_value)
                ,          order_by_direction   = substring_index(substring_index(v_label,',',find_in_set(id,v_value)),',',-1)
                where      find_in_set(id,v_value);
            when 'set join condition' then
                select group_concat(column_id1,comparison_operator,column_id2)
                into   v_label_alias
                from   qwz_join_condition
                where  find_in_set(concat(column_id1,comparison_operator,column_id2),p_choice)
                or     find_in_set(concat(column_id2,comparison_operator,column_id1),p_choice);

                insert into qwz_join_condition(column_id1, column_id2, comparison_operator)
                select      least(cast(mc1.value as unsigned), cast(mc2.value as unsigned))
                ,           greatest(cast(mc1.value as unsigned), cast(mc2.value as unsigned))
                ,           op.symbol
                from        qwz_menu_column1 mc1
                inner join  qwz_menu_column2 mc2
                on          mc1.position != mc2.position
                cross join (select '=' as symbol  union all
                            select '!='  union all
                            select '<=>' union all
                            select '>'   union all
                            select '<'   union all
                            select '>='  union all
                            select '<='
                           ) op
                where       find_in_set(concat(mc1.value,op.symbol,mc2.value),p_choice)
                on duplicate key update id = id, comparison_operator = op.symbol;

                delete from qwz_join_condition
                where  find_in_set(concat(column_id1,comparison_operator,column_id2),v_label_alias);

                set v_label_alias := 'Build SELECT list: + columns are added / - columns are removed';
                set v_menu_groups := 'columns,choose related tables,choose custom join,choose show sql,choose explain sql,choose execute sql,new query';

                set @qwz_dirty1:=true;
            when 'set related table' then

                set v_label_alias := 'Join path: + tables are added / - tables are removed';
                set v_menu_groups := 'related tables,choose custom join,choose columns,new query';

            when 'set schema' then
                set v_label_alias := 'Choose a table to start a join  path';                
                set @qwz_schema := v_value;
                if exists (
                    select null
                    from   qwz_table
                    limit  1
                ) then 
                    set @qwz_state := 'related tables';
                    set v_menu_groups := 'related tables,choose custom join,choose schema';
                else 
                    set @qwz_state := 'tables';
                    set v_menu_groups := 'choose schema,tables';
                end if;

            when 'set table' then
                set v_label_alias := 'Join path: + tables are added / - tables are removed';

                select coalesce(max(right_id),0) + 1
                into   v_left_id
                from   qwz_table;

                set v_right_id := v_left_id + 1;
                set v_schema_name := @qwz_schema;
                set v_table_name := v_label;

                insert into 
                qwz_table (  schema_name,  table_name,  left_id,  right_id, included, join_type)
                values    (v_schema_name,v_table_name,v_left_id,v_right_id,    'NO'
                ,         if(v_value='',NULL,substring_index(v_value,' ',1)));

                set v_menu_groups := 'related tables,choose custom join,choose columns,new query';

                set v_value := last_insert_id();

            when 'show sql' then
                set v_menu_groups := 'show sql,choose execute sql,choose explain sql,choose related tables,choose custom join,choose columns,new query';
            when 'tables' then
                set v_label_alias := 'Choose a table to start a join path.';
                set v_menu_groups := 'tables,choose schema,new query';
            else select 'case not found1', @qwz_state;
        end case;

        if @qwz_state in ('set table','set related table') then
            begin
                declare continue handler for not found 
                    set v_done := TRUE;
                insert into qwz_column(table_id, column_name
                ,          ordinal_position, included)
                select     t.id, column_name
                ,          ordinal_position, 'NO'
                from       qwz_table                  t
                inner join information_schema.columns c
                on         t.schema_name = c.table_schema
                and        t.table_name  = c.table_name
                where      find_in_set(t.id,v_value)
                and        t.included = 'NO'
                order by   c.ordinal_position;

                open csr_tables;
                _loop_tables: loop
                    fetch csr_tables 
                    into  v_id
                    ,     v_schema_name, v_table_name
                    ,     v_join_type, v_included;
                    if v_done then
                        set v_done := FALSE;
                        close csr_tables;
                        leave _loop_tables;
                    end if;

                    select left_id, right_id
                    into   v_left_id, v_right_id
                    from   qwz_table 
                    where  id = v_id;

                    if v_included = 'YES' then

                        delete     t.* , c.*, jc.*
                        from       qwz_table t
                        left join  qwz_column c
                        on         t.id = c.table_id
                        left join  qwz_join_condition jc
                        on         c.id = jc.column_id1
                        or         c.id = jc.column_id2
                        where      t.left_id   > v_left_id
                        and        t.right_id  < v_right_id;

                        update     qwz_table
                        set        included = if( left_id=v_left_id
                                              ,   'NO'
                                              ,   included)
                        ,          right_id = right_id - (v_right_id - v_left_id -1)
                        ,          left_id  = if( left_id > v_left_id
                                              ,   left_id - (v_right_id - v_left_id -1)
                                              ,   left_id)
                        where      right_id >= v_right_id;

                    else
                        select     count(distinct constraint_name) * 2
                        into       v_count_related_tables 
                        from       information_schema.key_column_usage kcu
                        where      kcu.table_schema = v_schema_name
                        and        kcu.table_name   = v_table_name
                        and        kcu.referenced_table_schema is not null;

                        update     qwz_table
                        set        left_id   = if( left_id > v_left_id
                                               ,   left_id + v_count_related_tables
                                               ,   left_id)
                        ,          right_id  = if( right_id >= v_right_id
                                               ,   right_id + v_count_related_tables
                                               ,   right_id)
                        ,          included  = if(id=v_id, 'YES', included)
                        where      right_id >= v_right_id
                        order by   left_id desc;

                        set @qwz_left_id := v_left_id - 1;
                        set @qwz_right_id := v_left_id;

                        insert into 
                        qwz_table (left_id, right_id
                        ,          schema_name, table_name
                        ,          constraint_schema, constraint_name
                        ,          included, join_type)
                        select     @qwz_left_id:=@qwz_left_id+2   left_id
                        ,          @qwz_right_id:=@qwz_right_id+2 right_id
                        ,          referenced_table_schema        schema_name
                        ,          referenced_table_name          table_name
                        ,          constraint_schema              constraint_scema
                        ,          constraint_name                constraint_name
                        ,          'NO'                           included
                        ,          if( max(cms.is_nullable)='NO'
                                   ,   coalesce(v_join_type,'INNER')
                                   ,   'LEFT'
                                   ) join_type
                        from       information_schema.key_column_usage kcu
                        inner join information_schema.columns          cms
                        on         kcu.table_schema = v_schema_name
                        and        kcu.table_name   = v_table_name
                        and        kcu.referenced_table_schema is not null
                        and        kcu.table_schema = cms.table_schema
                        and        kcu.table_name   = cms.table_name
                        and        kcu.column_name  = cms.column_name
                        group by   kcu.table_schema
                        ,          kcu.table_name
                        ,          kcu.constraint_name;
                    end if;

                end loop _loop_tables;

                truncate table qwz_table_ancestor;
                insert into qwz_table_ancestor(left_id, right_id)
                select left_id, right_id from qwz_table;

                set @qwz_dirty1 := true;
            end;
            set @qwz_state := 'related tables';
        end if;

        -- Filling the menu table
        _create_menu: begin
            declare v_menu_group_separator varchar(255) 
                default repeat('-',greatest(character_length(v_label_alias),64));
            declare v_add_menu_separator bool default false;
            declare v_index int unsigned;

            loop_menu_groups: loop

                if v_add_menu_separator then
                    -- add a separator between menu groups
                    insert into qwz_menu (id, position, label, state, value)
                    values (@qwz_id:=@qwz_id+1, '---',v_menu_group_separator, '', '');
                else
                    set v_add_menu_separator := true;
                end if;

                -- find an individual menu group amongst all groups
                set v_index := instr(v_menu_groups,',');
                if v_index = 0 then
                    set v_menu_group  := v_menu_groups;
                else 
                    set v_menu_group  := substring(v_menu_groups, 1, v_index - 1);
                    set v_menu_groups := substring(v_menu_groups, v_index+1);
                end if;

                -- handle the found menu group
                case v_menu_group
                    when 'choose columns' then
                        insert into qwz_menu (id, position, label, state, value)
                        values(@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,      'Choose columns', 'columns', '');
                    when 'choose custom join' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Add a table to CROSS JOIN', 'custom join', 'cross join')
                        ,      (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Add a table to INNER JOIN', 'custom join', 'inner join')
                        ,      (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Add a table to LEFT JOIN', 'custom join', 'left join')
                        ,      (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Add a table to RIGHT JOIN', 'custom join', 'right join');
                    when 'choose help' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Help', 'help', '');
                    when 'choose execute sql' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Execute SQL', 'execute sql', '');
                    when 'choose explain sql' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Explain SQL', 'explain sql', '');
                    when 'choose related tables' then
                        insert into qwz_menu (id, position, label, state, value)
                        values(@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,      'Add another table', 'related tables', '');
                    when 'choose schema' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,      concat('Set the schema (current: ', @qwz_schema, ')')
                        ,      'schemata', '');
                    when 'choose show sql' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'Show SQL', 'show sql', '');
                    when 'choose tables' then
                        insert into qwz_menu (id, position, label, state, value)
                        select @qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,      'Add a table', 'tables', ''
                        from   information_schema.tables
                        where  table_schema = @qwz_schema
                        having count(*) > 0;
                    when 'columns' then
                        insert into qwz_menu (id, position, label, state, value)
                        select     @qwz_id:=@qwz_id+1
                        ,          if( value=''
                                   ,   ''
                                   ,   @qwz_pos:=@qwz_pos+1)
                        ,          label
                        ,          state
                        ,          value
                        from       (
                        select     concat( 
                                       repeat('  ',count(distinct a.left_id)-1)
                                   ,   case 
                                           when type='table' then 
                                               concat(
                                                   c.schema_name, '.', c.table_name, '.*'
                                               ,   if(d.constraint_name is null
                                                   ,   ''
                                                   , concat(' (', d.constraint_name, ')'))
                                               )
                                           when type in(
                                               'table-overline'
                                           ,   'table-underline'
                                           ) then
                                               repeat(
                                                   '-'
                                               ,   character_length(c.schema_name)
                                               +   character_length(c.table_name)
                                               +   if( d.constraint_name is null
                                                   ,   0
                                                   ,   character_length(d.constraint_name)
                                                   +   3
                                                   )
                                               +   3
                                               )
                                           when type='column' then
                                               concat(
                                                   if( c.included='YES'
                                                   ,   concat(lpad(c.position_in_select,3,' '),'.')
                                                   ,   '+   ')
                                               ,   ' '
                                               ,   c.column_name
                                               )
                                       end
                                   )   label
                        ,          state
                        ,          coalesce(c.id,'')         value
                        from       qwz_table d
                        inner join qwz_table_ancestor a
                        on         d.left_id between a.left_id and a.right_id
                        inner join (
                                   select 'table-overline'    type
                                   ,      null                id
                                   ,      null                table_id
                                   ,      table_schema        schema_name
                                   ,      table_name
                                   ,      1000                ordinal_position
                                   ,      ''                  column_name
                                   ,      null                included
                                   ,      ''                  state
                                   ,      null                position_in_select
                                   from   information_schema.tables
                                   union all
                                   select 'table'
                                   ,      null
                                   ,      null
                                   ,      table_schema
                                   ,      table_name
                                   ,      -1
                                   ,      ''
                                   ,      null
                                   ,      ''                  state
                                   ,      null                position_in_select
                                   from   information_schema.tables
                                   union all
                                   select 'table-underline'       type
                                   ,      null
                                   ,      null
                                   ,      table_schema  schema_name
                                   ,      table_name
                                   ,      0             ordinal_position
                                   ,      ''            column_name
                                   ,      null
                                   ,      ''                  state
                                   ,      null                position_in_select
                                   from   information_schema.tables
                                   union all
                                   select 'column'
                                   ,      id
                                   ,      table_id
                                   ,      null
                                   ,      null
                                   ,      ordinal_position
                                   ,      column_name
                                   ,      included
                                   ,      'set column'        state
                                   ,      position_in_select
                                   from   qwz_column
                                   ) as c
                        on        (c.table_id is null
                        and        d.schema_name = c.schema_name
                        and        d.table_name  = c.table_name)
                        or         d.id          = c.table_id
                        where      d.included = 'YES'
                        group by   d.left_id
                        ,          ordinal_position
                        ) as a;

                        -- this delete is to remove the last table separator
                        delete from qwz_menu where id = @qwz_id;

                        -- we need copies to generate join conditions, order by columns etc
                        truncate table qwz_menu_column1;
                        insert into qwz_menu_column1(position,value)
                        select      m.position,m.value
                        from        qwz_menu m
                        inner join  qwz_column c
                        on          m.state = 'set column'
                        and         m.value = c.id
                        inner join  qwz_table t
                        on          c.table_id = t.id
                        and         t.included = 'YES';

                        -- add entries for order by columns
                        insert into qwz_menu(id, position, label, state, value, visible)
                        select     @qwz_id:=@qwz_id+1
                        ,          concat(mc1.position,d.suffix)
                        ,          d.label
                        ,          'set order by column'
                        ,          mc1.value
                        ,          'NO'
                        from       qwz_menu_column1 mc1
                        cross join(select 'A' as suffix, 'ASC' as label
                        union all  select 'D' as suffix, 'DESC' as label) d;

                        truncate table qwz_menu_column2;
                        insert into qwz_menu_column2
                        select      cm.position,cm.value
                        from        qwz_menu_column1 cm
                        inner join  qwz_column       c
                        on          cm.value = c.id
                        inner join  qwz_table t
                        on          c.table_id = t.id
                        and         t.constraint_name is null
                        and         t.join_type != 'CROSS';

                        -- generating join conditions
                        insert into qwz_menu (id, position, label, state, value, visible)
                        select     @qwz_id:=@qwz_id+1
                        ,          concat(mc1.position,op.symbol,mc2.position)
                        ,          op.symbol
                        ,          'set join condition'
                        ,          concat(least(mc1.value,mc2.value),',',greatest(mc1.value,mc2.value))
                        ,          'NO'
                        from       qwz_menu_column1 mc1
                        inner join qwz_menu_column2 mc2
                        on         mc1.position != mc2.position
                        cross join(select '=' as symbol  union all
                                   select '!='  union all
                                   select '<=>' union all
                                   select '>'   union all
                                   select '<'   union all
                                   select '>='  union all
                                   select '<='
                                  ) op
                        ;
                    when 'help' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, '', 'QWZ - A command-line query wizard for MySQL', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', '-------------------------------------------', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Welcome to QWZ! QWZ is a query wizard designed to be used from within the MySQL monitor,', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', 'also known as the MySQL command line client (the "mysql" executable.)', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', '', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', 'The query wizard is menu driven. Numbered choices appear as a list on screen. To select', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', 'an option, type:', '', '')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'CALL qwz(''<option>'');','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'In some cases, multiple options are allowed. Separate multiple options with a comma:','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'CALL qwz(''<option1>, <option2>,...,<optionN>'');','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Multiple options are allowed only if they are from the same menu group. A Menu group','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'is formed by a number of item appearing consecutively. Menu groups are separated from','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'one another by a horizontal "line" (that is, a stretch of dashes).','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Set the schema','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '-----------------','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'This option allows you to set the default schema. This schema is used to offer a choice','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'of tables to be added to the query','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Choose a table to start a join  path','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '------------------------------------','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Type any single number to select the table upon which the query is to be based.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Build join path','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '---------------','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'The list of query tables appears in the menu. If any of the query tables has','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'FOREIGN KEY constraints, then the tables referenced by the FOREIGN KEY constraint are','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'listed beneath that table. Related tables that have a + (plus sign) can be added to the','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'query, thereby lengthening or extending a join path. Once these tables are added,','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'a - (minus sign) appears and by selecting the table again it is taken out of the query','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '(in which case it get a + again, remaining available to be included at a later time).','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'When adding related tables, qwz automatically maintains the proper join condition.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'If the referencing columns are nullable, an outer join is automatically chosen.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'If a new table related to an outer-joined table is added, then that table is outer joined','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'too, regardless of the particular nullability of referencing columns.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Multiple related tables can be added to and removed from the query by specifying','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'a comma-separated list of the relevant menu options.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Custom joins','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '------------','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Tables that are not directly linked through a FOREIGN KEY constraint can also be added','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'by choosing one of the custom JOIN options (CROSS, INNER, LEFT, RIGHT). In this case,','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'you will usually have to set the exact join condition yourself, which is done in the ,','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Columns menu.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'In the Columns menu, join conditions are specified by typing: ','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '<column-number> <join-operator> <column-number>','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'where <column-number> is the number of the menu option that lists the column and ','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '<join-operator> is one of =, !=, >, <, >=, <= or <=>. Multiple join conditions may','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'thus be entered as a comma separated list.','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Columns','','')
                        ,      (@qwz_id:=@qwz_id+1, '', '-------','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'All columns of all tables included in the query are automatically available in the ','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'Columns menu. The columns with a + (plus sign) can be added to the query. For the ','','')
                        ,      (@qwz_id:=@qwz_id+1, '', 'columns that are already included in the query the ordinal position is listed instead.','','')
                        ;
                        set @qwz_state := 'start';
                    when 'new query' then
                        insert into qwz_menu (id, position, label, state, value)
                        values (@qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                               ,'New Query', 'start', '');
                    when 'related tables' then
                        insert into qwz_menu (id, position, label, state, value)
                        select     @qwz_id:=@qwz_id+1
                        ,          if( value=1
                                   ,   ''
                                   ,   @qwz_pos:=@qwz_pos+1)
                        ,          label
                        ,          state
                        ,          value
                        from       (
                        select     concat( 
                                       repeat('  ',count(a.left_id)-1)
                                   ,   case 
                                           when d.left_id = 1    then '  '
                                           when d.included='YES' then '- '
                                           when d.included='NO'  then '+ '
                                       end
                                   ,   d.schema_name
                                   ,   '.'
                                   ,   d.table_name
                                   ,   if(d.join_type is null, ''
                                       ,   concat(
                                               ' ('
                                           ,   d.join_type
                                           ,   if( d.constraint_name is null
                                               ,   ''
                                               ,   concat(' using FOREIGN KEY '
                                                   ,   d.constraint_name)
                                               )
                                           ,   ')'
                                           )
                                       )
                                   )                   label
                        ,          'set related table' state
                        ,          d.id                value
                        from       qwz_table d
                        inner join qwz_table_ancestor a
                        on         d.left_id between a.left_id and a.right_id
                        group by   d.left_id
                        ) as a;
                    when 'schemata' then
                        insert into qwz_menu (id, position, label, state, value)
                        select    @qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,         concat(
                                      if(schema_name=@qwz_schema,'* ','  ')
                                  ,   rpad(schema_name,64,' ')
                                  ,   ' (', count(table_name), ' tables)'
                                  )
                        ,         'set schema', schema_name
                        from      information_schema.schemata s
                        left join information_schema.tables   t
                        on        s.schema_name = t.table_schema
                        group by  s.schema_name
                        having    count(table_name) > 0;
                    when 'show sql' then
                        select @qwz_sql as "SQL";
                        set v_add_menu_separator:=false;
                    when 'tables' then
                        insert into qwz_menu (id, position, label, state, value)
                        select @qwz_id:=@qwz_id+1, @qwz_pos:=@qwz_pos+1
                        ,      table_name
                        ,      'set table'
                        ,      if(@qwz_state='custom join',v_value,'')
                        from   information_schema.tables
                        where  table_schema = @qwz_schema;

                    else 
                        select 'case not found2', v_menu_group;
                        leave _main;
                end case;

                if v_index = 0 then
                    leave loop_menu_groups;
                end if;
            end loop;
        end;

    end _real;
    -- Show the menu
    -- We need to use dynamic SQL to set the headings of the table.
    set @qwz_menu_stmt := concat(
                                  'select position as "", label as "'
                             ,    v_label_alias
                             ,    '" from qwz_menu '
                             ,    'where visible = ''YES'' '
                             ,    'order by id'
                             );
    prepare qwz_menu_stmt from @qwz_menu_stmt;
    execute qwz_menu_stmt;
    deallocate prepare qwz_menu_stmt;
end
//

delimiter ;

set @@sql_mode := @old_sql_mode;

call qwz('');